@@ -207,8 +207,7 @@ void imap_logout_all (void)
 }
 
 /* imap_read_literal: read bytes bytes from server into file. Not explicitly
- *   buffered, relies on FILE buffering. NOTE: strips \r from \r\n.
- *   Apparently even literals use \r\n-terminated strings ?! */
+ *   buffered, relies on FILE buffering. */
 int imap_read_literal (FILE* fp, IMAP_DATA* idata, unsigned int bytes, progress_t* pbar)
 {
   unsigned int pos;
@@ -228,7 +227,8 @@ int imap_read_literal (FILE* fp, IMAP_DATA* idata, unsigned int bytes, progress_
       return -1;
     }
 
-#if 1
+    /* Strip \r from \r\n, apparantly even literals use \r\n-terminated
+      strings ?! */
     if (r == 1 && c != '\n')
       fputc ('\r', fp);
 
@@ -239,7 +239,7 @@ int imap_read_literal (FILE* fp, IMAP_DATA* idata, unsigned int bytes, progress_
     }
     else
       r = 0;
-#endif
+
     fputc (c, fp);
 
     if (pbar && !(pos % 1024))
@@ -536,8 +536,9 @@ int imap_open_connection (IMAP_DATA* idata)
      * attacks is via $ssl_force_tls: an attacker can easily spoof
      * "* OK" and strip the STARTTLS capability.  So consult
      * $ssl_force_tls, not $ssl_starttls, to decide whether to
-     * abort. */
-    if (!idata->conn->ssf && !Tunnel && option(OPTSSLFORCETLS))
+     * abort. Note that if using $tunnel and $tunnel_is_secure,
+     * conn->ssf will be set to 1. */
+    if (!idata->conn->ssf && option(OPTSSLFORCETLS))
     {
       mutt_error _("Encrypted connection unavailable");
       mutt_sleep (1);
@@ -578,6 +579,143 @@ void imap_close_connection(IMAP_DATA* idata)
   memset (idata->cmds, 0, sizeof (IMAP_COMMAND) * idata->cmdslots);
 }
 
+/* Try to reconnect and merge current state back in.
+ * This is only done currently during mx_check_mailbox() polling
+ * when reopen is allowed. */
+int imap_reconnect (IMAP_DATA **p_idata)
+{
+  CONTEXT *orig_ctx, new_ctx;
+  int rc = -1, i;
+  IMAP_DATA *idata = *p_idata;
+  HEADER *old_hdr, *new_hdr;
+
+  /* L10N:
+     Message displayed when IMAP connection is lost and Mutt
+     tries to reconnect.
+  */
+  mutt_message _("Trying to reconnect...");
+  mutt_sleep (0);
+
+  orig_ctx = idata->ctx;
+  if (!orig_ctx)
+    goto cleanup;
+
+  if (mx_open_mailbox (orig_ctx->path,
+                       orig_ctx->readonly ? MUTT_READONLY : 0,
+                       &new_ctx) == NULL)
+    goto cleanup;
+
+  new_ctx.dontwrite = orig_ctx->dontwrite;
+  new_ctx.pattern = orig_ctx->pattern;
+  new_ctx.limit_pattern = orig_ctx->limit_pattern;
+
+  orig_ctx->pattern = NULL;
+  orig_ctx->limit_pattern = NULL;
+
+  if (idata->uid_validity == ((IMAP_DATA *) new_ctx.data)->uid_validity)
+  {
+    for (i = 0; i < new_ctx.msgcount; i++)
+    {
+      new_hdr = new_ctx.hdrs[i];
+      old_hdr = (HEADER *) int_hash_find (idata->uid_hash,
+                                          HEADER_DATA(new_hdr)->uid);
+      if (!old_hdr)
+        continue;
+
+      /* this logic is in part from mbox.c. */
+      if (old_hdr->changed)
+      {
+        mutt_set_flag (&new_ctx, new_hdr, MUTT_FLAG, old_hdr->flagged);
+        mutt_set_flag (&new_ctx, new_hdr, MUTT_REPLIED, old_hdr->replied);
+        mutt_set_flag (&new_ctx, new_hdr, MUTT_OLD, old_hdr->old);
+        mutt_set_flag (&new_ctx, new_hdr, MUTT_READ, old_hdr->read);
+
+        /* TODO: the ->env check is unfortunately voodoo that I
+         * haven't taken the time to track down yet.  It's in other
+         * parts of the code but I don't know why yet. */
+        if (old_hdr->env && old_hdr->env->changed)
+        {
+          new_hdr->env->changed = old_hdr->env->changed;
+          new_hdr->changed = 1;
+          new_ctx.changed = 1;
+
+          if (old_hdr->env->changed & MUTT_ENV_CHANGED_IRT)
+          {
+            mutt_free_list (&new_hdr->env->in_reply_to);
+            new_hdr->env->in_reply_to = old_hdr->env->in_reply_to;
+            old_hdr->env->in_reply_to = NULL;
+          }
+          if (old_hdr->env->changed & MUTT_ENV_CHANGED_REFS)
+          {
+            mutt_free_list (&new_hdr->env->references);
+            new_hdr->env->references = old_hdr->env->references;
+            old_hdr->env->references = NULL;
+          }
+          if (old_hdr->env->changed & MUTT_ENV_CHANGED_XLABEL)
+          {
+            FREE (&new_hdr->env->x_label);
+            new_hdr->env->x_label = old_hdr->env->x_label;
+            old_hdr->env->x_label = NULL;
+          }
+          if (old_hdr->env->changed & MUTT_ENV_CHANGED_SUBJECT)
+          {
+            FREE (&new_hdr->env->subject);
+            new_hdr->env->subject = old_hdr->env->subject;
+            new_hdr->env->real_subj = old_hdr->env->real_subj;
+            old_hdr->env->subject = old_hdr->env->real_subj = NULL;
+          }
+        }
+
+        if (old_hdr->attach_del)
+        {
+          if (old_hdr->content->parts && !new_hdr->content->parts)
+          {
+            new_hdr->attach_del = 1;
+            new_hdr->changed = 1;
+            new_ctx.changed = 1;
+            new_hdr->content->parts = old_hdr->content->parts;
+            old_hdr->content->parts = NULL;
+          }
+        }
+      }
+
+      mutt_set_flag (&new_ctx, new_hdr, MUTT_DELETE, old_hdr->deleted);
+      mutt_set_flag (&new_ctx, new_hdr, MUTT_PURGE, old_hdr->purge);
+      mutt_set_flag (&new_ctx, new_hdr, MUTT_TAG, old_hdr->tagged);
+    }
+  }
+
+  rc = 0;
+
+cleanup:
+  idata->status = IMAP_FATAL;
+  mx_fastclose_mailbox (orig_ctx);
+  imap_close_connection (idata);
+
+  if (rc != 0)
+  {
+    /* L10N:
+       Message when Mutt tries to reconnect to an IMAP mailbox but is
+       unable to.
+    */
+    mutt_error _("Reconnect failed.  Mailbox closed.");
+  }
+  else
+  {
+    memcpy (orig_ctx, &new_ctx, sizeof(CONTEXT));
+    idata = (IMAP_DATA *)orig_ctx->data;
+    idata->ctx = orig_ctx;
+    *p_idata = idata;
+    /* L10N:
+       Message when Mutt reconnects to an IMAP mailbox after a fatal error.
+    */
+    mutt_error _("Reconnect succeeded.");
+  }
+  mutt_sleep (0);
+
+  return rc;
+}
+
 /* imap_get_flags: Make a simple list out of a FLAGS response.
  *   return stream following FLAGS response */
 static char* imap_get_flags (LIST** hflags, char* s)
@@ -1625,7 +1763,7 @@ int imap_check_mailbox (CONTEXT *ctx, int *index_hint, int force)
   /* overload keyboard timeout to avoid many mailbox checks in a row.
    * Most users don't like having to wait exactly when they press a key. */
   IMAP_DATA* idata;
-  int result = 0;
+  int result = -1, poll_rc;
 
   idata = (IMAP_DATA*) ctx->data;
 
@@ -1634,19 +1772,19 @@ int imap_check_mailbox (CONTEXT *ctx, int *index_hint, int force)
       && (idata->state != IMAP_IDLE || time(NULL) >= idata->lastread + ImapKeepalive))
   {
     if (imap_cmd_idle (idata) < 0)
-      return -1;
+      goto errcleanup;
   }
   if (idata->state == IMAP_IDLE)
   {
-    while ((result = mutt_socket_poll (idata->conn, 0)) > 0)
+    while ((poll_rc = mutt_socket_poll (idata->conn, 0)) > 0)
     {
       if (imap_cmd_step (idata) != IMAP_CMD_CONTINUE)
       {
         dprint (1, (debugfile, "Error reading IDLE response\n"));
-        return -1;
+        goto errcleanup;
       }
     }
-    if (result < 0)
+    if (poll_rc < 0)
     {
       dprint (1, (debugfile, "Poll failed, disabling IDLE\n"));
       mutt_bit_unset (idata->capabilities, IDLE);
@@ -1656,12 +1794,31 @@ int imap_check_mailbox (CONTEXT *ctx, int *index_hint, int force)
   if ((force ||
        (idata->state != IMAP_IDLE && time(NULL) >= idata->lastread + Timeout))
       && imap_exec (idata, "NOOP", IMAP_CMD_POLL) != 0)
-    return -1;
+    goto errcleanup;
 
   /* We call this even when we haven't run NOOP in case we have pending
    * changes to process, since we can reopen here. */
   imap_cmd_finish (idata);
 
+  result = 0;
+
+errcleanup:
+  /* Try to reconnect Context if a cmd_handle_fatal() was flagged */
+  if (idata->status == IMAP_FATAL)
+  {
+    if ((idata->reopen & IMAP_REOPEN_ALLOW) &&
+        Context &&
+        idata->ctx == Context)
+    {
+      if (imap_reconnect (&idata) == 0)
+      {
+        idata->check_status = 0;
+        return MUTT_RECONNECTED;
+      }
+    }
+    return -1;
+  }
+
   if (idata->check_status & IMAP_EXPUNGE_PENDING)
     result = MUTT_REOPENED;
   else if (idata->check_status & IMAP_NEWMAIL_PENDING)
@@ -2110,8 +2267,6 @@ int imap_subscribe (char *path, int subscribe)
   IMAP_DATA *idata;
   char buf[LONG_STRING*2];
   char mbox[LONG_STRING];
-  int mblen;
-  BUFFER err;
   IMAP_MBOX mx;
 
   if (!mx_is_imap (path) || imap_parse_path (path, &mx) || !mx.mbox)
@@ -2128,16 +2283,10 @@ int imap_subscribe (char *path, int subscribe)
 
   if (option (OPTIMAPCHECKSUBSCRIBED))
   {
-    mutt_buffer_init (&err);
-    err.dsize = STRING;
-    err.data = safe_malloc (err.dsize);
-    mblen = snprintf (mbox, sizeof (mbox), "%smailboxes ",
-                      subscribe ? "" : "un");
-    imap_quote_string_and_backquotes (mbox + mblen, sizeof(mbox) - mblen,
-                                      path);
-    if (mutt_parse_rc_line (mbox, &err))
-      dprint (1, (debugfile, "Error adding subscribed mailbox: %s\n", err.data));
-    FREE (&err.data);
+    if (subscribe)
+      mutt_buffy_add (path, NULL, -1);
+    else
+      mutt_buffy_remove (path);
   }
 
   if (subscribe)
